/* 
 *  socket interface fuer shell scripts
 * Fuer vi:    set ts=4 sw=4 (Tabs auf 4 stellen)
 *
 * Angepasst fr Linux getestet unter Suse 6.4 
 * Frank Meyer @ Wiesemann & Theis GmbH  22.01.2001
 */

// static char     RCSID[] = "$Header: /u/guenter/theis/socket/RCS/sockif.c,v 1.5 1994/05/05 11:53:54 guenter Exp $";
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <unistd.h> 
#include <getopt.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define STDIN   0               /* standard input */
#define STDOUT  1               /* standard output */
#define STDERR  2               /* standard error */

int      Socket;                /* Socket Nummer */
char    *Prgname;               /* Programmname */
char     Extprg[128];           /* externes Programm mit  * *
                                 * stdin,stdout <-> Socket */
int      getservice(char *num);
void     komm(void);
void     open_pipes(char *prog);
void     sighandler(int sig);
int      sock_open(char *remote, int port);
char    *sysmerr(char *msg);
int      transfer(int from, int to);
void     usage(void);

int      main(int argc, char **argv)
{
    extern int      optind;
    extern char    *optarg;
    extern char    *optstring;
    int             cc;
    int             i;
    int             port;              /* Portnummer */
    char            remote[128];       /* Name oder Internetadr. */
    char            service[128];      /* Sevicename oder Portnummer */

    Prgname = argv[0];
    Extprg[0] = '\0';

    while((cc = getopt(argc, argv, "e:h")) != -1)
    { switch (cc)
      { case 'e':                      /* Name des externen Programmes */
                 strncpy(Extprg, optarg, sizeof(Extprg));
                break;
        case 'h':                      /* help = usage Anzeige */
                 usage();
                 exit(-1);
                break;
    } }
    // Internetadresse und Portnummer angegeben ?
    if( argc - optind != 2 )
      { usage();
        exit(1);
      }
    strncpy( remote,  argv[optind],     sizeof(remote));
    strncpy( service, argv[optind + 1], sizeof(service));
    // Alle Signals, ausser SIGUSR1, beenden das Programm 
    // Eventuell noch weitere Signals abfangen, z.B. Jobcontrol etc.
    for( i = 0; i < NSIG; i++ )
       { if( i == SIGUSR1 )
             continue;
         signal( i, sighandler );
       }
    // Eventuell Servicename in Portnummer umwandeln
    port = getservice(service);
    if( port < 0 )
      { fprintf( stderr, "%s: Unbekannter Service\n", Prgname );
        exit( -1 );
      }
    // Socket oeffnen
    Socket = sock_open( remote, port );
    if( Socket == -1 )
      { fprintf( stderr, "%s\n", sysmerr( "Socket oeffnen" ));
        exit( errno );
      }
     else if( Socket == -2 )
      { fprintf( stderr, "%s: Unbekannte Internetadresse %s\n", Prgname, remote );
        exit( -1 );
      }
    // 
    // falls ein externes Programm aufgerufen wird, pipes zu stdin/stdout
    // dieses Programmes anlegen
    if( Extprg[0] != '\0' )
        open_pipes( Extprg );
    // Kommunikation stdin/stdout <-> Socket bearbeiten 
    komm();
    // Socket schliessen --> Programmende
    close( Socket );
}

/* 
 *    Socket zu Internetadresse remote, Port port oeffnen
 *
 *  Input:  char *remote, Inernetname oder Adresse
 *          int port, Portnummer
 *
 *  Return: int s, Socketnummer
 *          -1, Fehler
 */
int sock_open( char *remote, int port )
{   struct sockaddr_in sa;
    struct hostent *hp;
    int             a,
                    s;
    long            addr;

    bzero( &sa, sizeof(sa));
    //  Internetadresse ist als xxx.xxx.xxx.xxx angegeben
    if( (addr = inet_addr(remote)) != -1 )
      { bcopy(&addr, (char *)&sa.sin_addr, sizeof(addr));
        sa.sin_family = AF_INET;
      }
     // Internetadresse ist als FQDN angegeben, Aufloesung ueber Nameserver
     // bzw. /etc/hosts
     else
      { if((hp = gethostbyname(remote)) == NULL)
            return -2;
        bcopy(hp->h_addr, (char *)&sa.sin_addr, hp->h_length);
        sa.sin_family = hp->h_addrtype;
      }
    sa.sin_port = htons((u_short) port);
    // Oeffne Socket
    if(( s = socket( sa.sin_family, SOCK_STREAM, 0 )) < 0)
        return -1;
    // Connect Socket
    if( connect( s, &sa, sizeof(sa)) < 0 )
      { close( s );
        return -1;
      }
   return s;
}

/* 
 * Falls Portnummer ein Servicename war, wird die Portnummer ueber
 * /etc/services bestimmt.
 *
 *  Input:  char *num, string mit Portnummer oder Servicename
 *  Return: int port, Portnummer als Integer
 *                    bei Fehler -1
 */
int getservice( char *num )
{
    struct servent *servent;
    int             port;

    if((port = atoi(num)) == 0 )
      { servent = getservbyname( num, "tcp" );
        if( servent == NULL )
            port = -1;
         else
            port = ntohs(servent->s_port);
      }
   return port;
}

/* 
 *    Tranferiere Bytes von from nach to
 *
 *  Input: int from, Filedescriptor, von dem gelesen werden soll
 *         int to,   Filedescriptor, auf den geschrieben werden soll
 *
 *  Return:  0, OK
 *          -1, bei Fehler
 */
int transfer( int from, int to )
{   int             len;
    int             written;
    char           *ptr;
    char            buf[BUFSIZ];

    if((len = read( from, buf, BUFSIZ)) == -1 )
      { fprintf( stderr, "%s\n", sysmerr("transfer-read"));
        return -1;
      }
    if( len == 0 )                      /* End of File */
      { return -1;
      }

    // Bytes von Lesefd nach Schreibfd ausgeben
    ptr = buf;
    while( len > 0 )
    { written = write( to, ptr, len );
      if( written == -1 )
        { fprintf( stderr, "%s\n", sysmerr("transfer-write"));
          return -1;
        }
      len -= written;
      ptr += written;
    }
  return 0;
}

/* 
 *    Kommunikation zwischen stdin/stdout und Socket durchfuehren
 */
void komm( void )
{   fd_set  readfds;
       int  nfds;
       int  fd;

    // Anzahl fds fuer select call  setzen
    nfds = Socket + 1;
    // Schleife wird nur durch Fehler oder Signals abgebrochen
    while( 1 )
    { // Read fd's fuer select call setzen
      FD_ZERO(&readfds);
      FD_SET( STDIN, &readfds );
      FD_SET( Socket, &readfds);
      // warte auf Input von stdin oder socket
      fd = select( nfds, &readfds, NULL, NULL, NULL );
      if( fd < 0 )
        {                             /* Fehler oder Signal empfangen */
          fprintf(stderr, "%s\n", sysmerr("select call"));
          exit(-1);
        }
      // transferiere Bytes von readfd nach writefd
      if( FD_ISSET( Socket, &readfds ))     /* von Socket lesen ? */
        { if( transfer( Socket, STDOUT ) < 0)
            { break;
        }   }
       else // von stdin lesen 
        { if( transfer( STDIN, Socket ) < 0)
            { break;
    }   }   }
}

/* 
 *    stdin/stdout/stderr mit externem Programm verbinden
 *
 *  Input:  char *prog, Name des externen Programmes
 *
 */
void open_pipes( char *prog )
{
    int  from_fds[2];       // from child process 
    int  to_fds[2];         // to child process 

    // Pipes anlegen 
    if( pipe( from_fds ) == -1)
      { fprintf( stderr, "%s\n", sysmerr("pipe-from call"));
        exit( errno );
      }
     else if( pipe(to_fds) == -1 )
      { fprintf( stderr, "%s\n", sysmerr("pipe-to call"));
        exit(errno);
      }
    // Fd's fuer child Process umlegen
    switch( fork())
    { case 0: // child
              // stdin mit pipe verbinden
              close( STDIN );
              close( to_fds[1] );
               dup2( to_fds[0], STDIN );
              close( to_fds[0 ]);
              // stdout mit pipe verbinden
              close( STDOUT );
              close( from_fds[0] );
               dup2( from_fds[1], STDOUT );
              // stderr mit pipe verbinden
              close( STDERR );
               dup2( from_fds[1], STDERR );
              close( from_fds[1] );
              // externes Programm aufrufen
              execl( "/bin/sh", "sh", "-c", prog, NULL );
              // Fehler beim exec call
              fprintf( stderr, "%s\n", sysmerr("exec"));
              // parent process beenden
              kill( getppid(), SIGUSR1 );
            exit(255);

     case -1: // Fehler beim fork call
              fprintf( stderr, "%s\n", sysmerr("fork"));
              exit( errno );
     default: // parent process
              // stdin mit pipe verbinden
              close( STDIN );
              close( from_fds[1] );
               dup2( from_fds[0], STDIN );
              close( from_fds[0] );
              // stdout mit pipe verbinden
              close( STDOUT );
              close( to_fds[0] );
               dup2( to_fds[1], STDOUT );
              close( to_fds[1] );
    }
}

/* 
 *    Fehlermeldung mit Angabe Systemmeldung ausgeben
 *
 *  Input: char *msg, Usermeldung
 *  Return: char *buf, Usermeldung + Systemmeldung
 *
 */
char *sysmerr( char *msg )
{   static char     buf[256];
    int             lerrno;
    int             cnt;
    extern int      errno,
                    sys_nerr;
    extern __const char *__const sys_errlist[];

    lerrno = errno;
    cnt = 15;                          // max. Laenge von 'ERROR: ...'
    if( strlen( msg ) > 240)           // string zu lang 
      { sprintf( buf, "ERROR: (%d", lerrno );
      }
     else
      { sprintf( buf, "ERROR: %s (%d", msg, lerrno );
        cnt += strlen(msg);
      }
    if( lerrno > 0 & lerrno < sys_nerr )
      { if( strlen( sys_errlist[lerrno]) + cnt > 253 )    // string zu lang
            sprintf( &buf[strlen(buf)], ")" );
         else
            sprintf( &buf[strlen(buf)], "; %s)", sys_errlist[lerrno]);
      }
     else
      { sprintf( &buf[strlen(buf)], ")" );
      }
   return buf;
}

/* 
 *    Ausgabe Usage Meldung
 *
 */
void usage( void )
    {
     fprintf( stderr, "Usage: %s [-e prg] ipadresse portnummer\n", Prgname);
    }

/* 
 *    Signalroutine, Programmabbruch und Angabe des empfangenen Signals
 *
 *  Input: int sig, Signalnummer
 */
void sighandler( int sig )
    { fprintf( stderr, "\n%s: Signal %d empfangen, Programm beendet\n", Prgname, sig );
      exit( -sig );
    }
